"""
Abstract BetterUptime Resource
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Union

from yarl import URL

from betteruptime.api.http_client import HTTPClient


class AbstractResource(ABC):
    """
    Abstract BetterUptime Resource, all resources should inherit from
    this class.
    """

    _http_client: HTTPClient
    _name: str
    _resource_id: Optional[str] = None

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__()
        self.http_client = http_client

    @property
    def http_client(self) -> HTTPClient:
        """
        http_client property getter.
        """
        return self._http_client

    @http_client.setter
    def http_client(self, http_client: HTTPClient) -> None:
        """
        http_client property setter.
        """
        self._http_client = http_client

    @property
    def name(self) -> str:
        """
        name property getter.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """
        name property setter.
        """
        self._name = name

    @property
    def resource_id(self) -> Optional[str]:
        """
        resource_id property getter.
        """
        return self._resource_id

    @abstractmethod
    def __call__(self, resource_id: str) -> AbstractResource:
        pass

    def _get_base_path(self) -> URL:
        """
        returns resource's path.
        """
        return URL(self.name)


class AbstractSubResource(AbstractResource):
    """
    Abstract BetterUptime Sub-resource.
    """

    _parent: AbstractResource

    def __init__(self, http_client: HTTPClient, parent: AbstractResource) -> None:
        super().__init__(http_client=http_client)
        self.parent = parent

    @property
    def parent(self) -> Union[AbstractResource, AbstractSubResource]:
        """
        parent property getter.
        """
        return self._parent

    @parent.setter
    def parent(self, parent: Union[AbstractResource, AbstractSubResource]) -> None:
        """
        parent property setter.
        """
        self._parent = parent

    @abstractmethod
    def __call__(self, resource_id: str) -> AbstractSubResource:
        pass

    def _build_path(
        self,
        path: URL,
        parent: Union[AbstractResource, AbstractSubResource, None] = None,
    ) -> URL:
        if parent is None:
            path_part = self._build_path(path=URL(self.name), parent=self.parent)
            path = path_part / str(path)
        elif isinstance(parent, AbstractSubResource):
            path_part = self._build_path(path=path, parent=parent.parent)
            if parent.resource_id:
                path = path_part / str(URL(parent.name) / parent.resource_id)
            else:
                path = path_part / parent.name
        elif isinstance(parent, AbstractResource):
            if parent.resource_id is not None:
                path = URL(parent.name) / parent.resource_id

        return path

    def _get_base_path(self) -> URL:
        """
        returns resource's path.
        """
        return URL(self.name)



"""
BetterUptime Monitor Groups Resource
"""
from __future__ import annotations

from typing import Generator

from yarl import URL

from betteruptime.api.exceptions import ApiError
from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.generic import MutableResource
from betteruptime.typing import JSON
from betteruptime.util.errors import parse_error_response


class MonitorGroup(MutableResource):
    """
    Represents BetterUptime Monitor Groups Resource
    """

    def __init__(self, http_client: HTTPClient, name: str = "monitor-groups") -> None:
        super().__init__(http_client, name)

    def __call__(self, resource_id: str) -> MonitorGroup:
        new_resource = MonitorGroup(http_client=self.http_client)
        new_resource._resource_id = resource_id
        return new_resource

    def monitors(self, page: int = 1) -> JSON:
        """
        List paginated monitors in this group.
        """
        if self.resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.monitors."
                f" You must use {self.__class__.__name__}('12345').monitors."
            )

        result = self.http_client.get(
            path=(self._get_base_path() / self.resource_id / "monitors").update_query(page=page)
        )
        if 200 == result.status_code:
            payload: JSON = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def monitors_iter(self, page: int = 1) -> Generator[JSON, None, None]:
        """
        List all monitor items by itering over all pages.
        """
        while True:
            result = self.monitors(page=page)
            assert isinstance(result, dict)
            for monitor in result["data"]:
                yield monitor
            if result["pagination"]["next"]:
                next_url: URL = URL(result["pagination"]["next"])
                page = int(next_url.query["page"])
            else:
                break



"""
BetterUptime Monitors Resource
"""
from __future__ import annotations

from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.generic import ImmutableResource


class EscalationPolicy(ImmutableResource):
    """
    Represents BetterUptime Escalation Policies Resource.
    """

    def __init__(self, http_client: HTTPClient, name: str = "policies") -> None:
        super().__init__(http_client, name)

    def __call__(self, resource_id: str) -> EscalationPolicy:
        new_resource = EscalationPolicy(http_client=self.http_client)
        new_resource._resource_id = resource_id
        return new_resource



"""
BetterUptime Heartbeat Groups Resource
"""
from __future__ import annotations

# from betteruptime.api.exceptions import ApiError
from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.generic import MutableResource

# from typing import Any, Dict


class HeartbeatGroup(MutableResource):
    """
    Represents BetterUptime Heartbeat Groups Resource
    """

    def __init__(self, http_client: HTTPClient, name: str = "heartbeat-groups") -> None:
        super().__init__(http_client, name)

    def __call__(self, resource_id: str) -> HeartbeatGroup:
        new_resource = HeartbeatGroup(http_client=self.http_client)
        new_resource._resource_id = resource_id
        return new_resource

    # @property
    # def heartbeats(self) -> Dict[str, Any]:
    #     """
    #     List heartbeats in this group.
    #     """
    #     if self.resource_id is None:
    #         raise ValueError(
    #             f"A resource_id is mandatory to call {self.__class__.__name__}.heartbeats."
    #             f" You must use {self.__class__.__name__}('12345').heartbeats."
    #         )

    #     result = self.http_client.get(
    #         path=self._get_base_path() / self.resource_id / "heartbeats"
    #     )
    #     if 200 == result.status_code:
    #         return result.json()
    #     raise ApiError(
    #         resource=self.name, status_code=result.status_code, reason=result.reason
    #     )



"""
BetterUptime generic Resource classes
"""
from __future__ import annotations

from typing import Generator, Optional

from yarl import URL

from betteruptime.api.exceptions import ApiError
from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.abstract import AbstractResource, AbstractSubResource
from betteruptime.typing import JSON
from betteruptime.util.errors import parse_error_response


class ImmutableResource(AbstractResource):
    """
    Immutable BetterUptime Resource.
    """

    def __init__(self, http_client: HTTPClient, name: str) -> None:
        super().__init__(http_client)
        self.name = name

    def get(self, resource_id: Optional[str] = None) -> JSON:
        """
        Get a single resource.
        """
        resource_id = resource_id or self.resource_id
        if resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.get()."
                f" You can either use {self.__class__.__name__}.get('12345') or"
                f" {self.__class__.__name__}('12345').get()."
            )

        result = self.http_client.get(path=self._get_base_path() / resource_id)
        if 200 == result.status_code:
            payload: JSON = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def list(self, page: int = 1) -> JSON:
        """
        List paginated resource.
        """
        result = self.http_client.get(path=self._get_base_path().update_query(page=page))
        if 200 == result.status_code:
            payload: JSON = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def list_iter(self, page: int = 1) -> Generator[JSON, None, None]:
        """
        List all resource items by itering over all pages.
        """
        while True:
            result = self.list(page=page)
            assert isinstance(result, dict)
            for monitor in result["data"]:
                yield monitor
            if result["pagination"]["next"]:
                next_url: URL = URL(result["pagination"]["next"])
                page = int(next_url.query["page"])
            else:
                break


class ImmutableSubResource(AbstractSubResource):
    """
    Immutable BetterUptime SubResource.
    """

    def __init__(self, http_client: HTTPClient, parent: AbstractResource, name: str) -> None:
        super().__init__(http_client=http_client, parent=parent)
        self.name = name

    def get(self, resource_id: Optional[str] = None) -> JSON:
        """
        Get a single sub-resource.
        """
        resource_id = resource_id or self.resource_id
        if resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.get()."
                f" You can either use {self.__class__.__name__}.get('12345') or"
                f" {self.__class__.__name__}('12345').get()."
            )

        result = self.http_client.get(path=self._build_path(URL(self.name)) / resource_id)
        if 200 == result.status_code:
            payload: JSON = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def list(self, page: int = 1) -> JSON:
        """
        List paginated sub-resource.
        """
        path: URL = self._build_path(URL(self.name))
        result = self.http_client.get(path=path.update_query(page=page))
        if 200 == result.status_code:
            payload: JSON = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def list_iter(self, page: int = 1) -> Generator[JSON, None, None]:
        """
        List all sub-resource itmes by itering over all pages.
        """
        while True:
            result = self.list(page=page)
            assert isinstance(result, dict)
            for monitor in result["data"]:
                yield monitor
            if result["pagination"]["next"]:
                next_url: URL = URL(result["pagination"]["next"])
                page = int(next_url.query["page"])
            else:
                break


class MutableResource(ImmutableResource):
    """
    Mutable BetterUptime Resource.
    """

    def create(self, payload: JSON) -> JSON:
        """
        Create resource.
        """
        result = self.http_client.post(path=self._get_base_path(), json=payload)
        if 201 == result.status_code:
            payload = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def delete(self, resource_id: Optional[str] = None) -> JSON:
        """
        Delete resource.
        """
        resource_id = resource_id or self.resource_id
        if resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.delete()."
                f" You can either use {self.__class__.__name__}.delete('12345') or"
                f" {self.__class__.__name__}('12345').delete()."
            )

        result = self.http_client.delete(path=self._get_base_path() / resource_id)
        if 204 == result.status_code:
            return None

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def update(self, payload: JSON, resource_id: Optional[str] = None) -> JSON:
        """
        Update resource.
        """
        resource_id = resource_id or self.resource_id
        if resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.update()."
                f" You can either use {self.__class__.__name__}.update('12345') or"
                f" {self.__class__.__name__}('12345').update()."
            )

        result = self.http_client.patch(path=self._get_base_path() / resource_id, json=payload)
        if 200 == result.status_code:
            payload = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )


class MutableSubResource(ImmutableSubResource):
    """
    Mutable BetterUptime SubResource.
    """

    def create(self, payload: JSON) -> JSON:
        """
        Create resource.
        """
        result = self.http_client.post(path=self._build_path(URL(self.name)), json=payload)
        if 201 == result.status_code:
            payload = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def delete(self, resource_id: Optional[str] = None) -> JSON:
        """
        Delete resource.
        """
        resource_id = resource_id or self.resource_id
        if resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.delete()."
                f" You can either use {self.__class__.__name__}.delete('12345') or"
                f" {self.__class__.__name__}('12345').delete()."
            )

        result = self.http_client.delete(path=self._build_path(URL(self.name)) / resource_id)
        if 204 == result.status_code:
            return None

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def update(self, payload: JSON, resource_id: Optional[str] = None) -> JSON:
        """
        Update resource.
        """
        resource_id = resource_id or self.resource_id
        if resource_id is None:
            raise ValueError(
                f"A resource_id is mandatory to call {self.__class__.__name__}.update()."
                f" You can either use {self.__class__.__name__}.update('12345') or"
                f" {self.__class__.__name__}('12345').update()."
            )

        result = self.http_client.patch(
            path=self._build_path(URL(self.name)) / resource_id,
            json=payload,
        )
        if 200 == result.status_code:
            payload = result.json()
            return payload

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )



"""
BetterUptime Client.
"""
from __future__ import annotations

from . import resources
from .api import api_client, http_client
from .api.api_client import Client
from .version import version as __version__

__all__ = ["Client", "api_client", "http_client", "resources", "__version__"]



"""
BetterUptime format helpers.
"""
# 3rdp
from yarl import URL


def construct_url(url: URL, path: URL) -> str:
    """Helper to construct URL"""
    return f"{str(url).rstrip('/')}/{str(path).strip('/')}"



"""
BetterUptime API
"""
from typing import Dict, Optional

# API settings
_API_HOST: str = "https://betteruptime.com"
_API_MAX_RETRIES: int = 3
_API_PROXIES: Optional[Dict[str, str]] = None
_API_TIMEOUT: float = 30.0
_API_VERIFY: bool = True
_API_VERSION: str = "v2"



"""
BetterUptime Monitors Resource
"""
from __future__ import annotations

from http import HTTPStatus
from typing import Any

from betteruptime.api.exceptions import ApiError
from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.generic import MutableResource
from betteruptime.typing import JSON
from betteruptime.util.errors import parse_error_response


class Monitor(MutableResource):
    """
    Represents BetterUptime Monitors Resource
    """

    def __init__(self, http_client: HTTPClient, name: str = "monitors") -> None:
        super().__init__(http_client, name)

    def __call__(self, resource_id: str) -> Monitor:
        new_resource = Monitor(http_client=self.http_client)
        new_resource._resource_id = resource_id
        return new_resource

    def get_by_name(self, name: str) -> JSON:
        """
        Get a single monitor by name.
        """
        if name is None:
            raise ValueError(
                f"An url is mandatory to call {self.__class__.__name__}.get_by_name()."
                f" You must use {self.__class__.__name__}.get_by_name('Backend')."
            )

        result = self.http_client.get(path=self._get_base_path().update_query(pronounceable_name=name))
        if 200 == result.status_code:
            exists = result.json()
            if len(exists["data"]) == 1:
                return {"data": exists["data"][0]}
            raise ApiError(
                resource=self.name,
                status_code=HTTPStatus.NOT_FOUND,
                reason=HTTPStatus.NOT_FOUND.description,
                errors=None,
            )

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def get_by_url(self, url: str) -> JSON:
        """
        Get a single monitor by url.
        """
        if url is None:
            raise ValueError(
                f"An url is mandatory to call {self.__class__.__name__}.get_by_url()."
                f" You must use {self.__class__.__name__}.get_by_url('http://my.company')."
            )

        result = self.http_client.get(path=self._get_base_path().update_query(url=url))
        if 200 == result.status_code:
            exists = result.json()
            if len(exists["data"]) == 1:
                return {"data": exists["data"][0]}
            raise ApiError(
                resource=self.name,
                status_code=HTTPStatus.NOT_FOUND,
                reason=HTTPStatus.NOT_FOUND.description,
                errors=None,
            )

        raise ApiError(
            resource=self.name,
            status_code=result.status_code,
            reason=result.reason,
            errors=parse_error_response(result),
        )

    def delete_by_name(self, name: str) -> Any:
        """
        Delete a single monitor by name.
        """
        if name is None:
            raise ValueError(
                f"An url is mandatory to call {self.__class__.__name__}.delete_by_name()."
                f" You must use {self.__class__.__name__}.delete_by_name('Backend')."
            )

        monitor = self.get_by_name(name=name)
        assert isinstance(monitor, dict)
        self.delete(monitor["data"]["id"])

    def delete_by_url(self, url: str) -> Any:
        """
        Delete a single monitor by url.
        """
        if url is None:
            raise ValueError(
                f"An url is mandatory to call {self.__class__.__name__}.delete_by_url()."
                f" You must use {self.__class__.__name__}.delete_by_url('http://my.company')."
            )

        monitor = self.get_by_url(url=url)
        assert isinstance(monitor, dict)
        self.delete(monitor["data"]["id"])



"""
BetterUptime Status Pages Resource
"""
from __future__ import annotations

from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.generic import MutableResource

from .reports import StatusPageReport
from .resources import StatusPageResource
from .sections import StatusPageSection


class StatusPage(MutableResource):
    """
    Represents BetterUptime Status Page Resource.
    """

    _reports: StatusPageReport
    _resources: StatusPageResource
    _sections: StatusPageSection

    def __init__(self, http_client: HTTPClient, name: str = "status-pages") -> None:
        super().__init__(http_client, name)
        self.reports = StatusPageReport(http_client=http_client, parent=self)
        self.resources = StatusPageResource(http_client=http_client, parent=self)
        self.sections = StatusPageSection(http_client=http_client, parent=self)

    def __call__(self, resource_id: str) -> StatusPage:
        new_resource = StatusPage(http_client=self.http_client)
        new_resource._resource_id = resource_id
        return new_resource

    @property
    def reports(self) -> StatusPageReport:
        """
        reports property getter.
        """
        return self._reports

    @reports.setter
    def reports(self, reports: StatusPageReport) -> None:
        """
        reports property setter.
        """
        self._reports = reports

    @property
    def resources(self) -> StatusPageResource:
        """
        resources property getter.
        """
        return self._resources

    @resources.setter
    def resources(self, resources: StatusPageResource) -> None:
        """
        resources property setter.
        """
        self._resources = resources

    @property
    def sections(self) -> StatusPageSection:
        """
        sections property getter.
        """
        return self._sections

    @sections.setter
    def sections(self, sections: StatusPageSection) -> None:
        """
        sections property setter.
        """
        self._sections = sections



"""
pytest configuration
"""
from typing import Any, Dict

import pytest

import betteruptime


@pytest.fixture()
def client() -> betteruptime.Client:
    """
    BetterUptime Client instance.
    """
    return betteruptime.Client(bearer_token="fake")


def clean_response(response):
    """Remove some info from the response before writing cassettes."""
    remove_headers = {
        "Set-Cookie",
        "set-cookie",
        "Date",
        "x-request-id",
        "x-runtime",
        "alt-svc",
        "NEL",
        "Expect-CT",
        "Report-To",
    }
    if isinstance(response["headers"], dict):
        # Normal client stores headers as dict
        for header_name in remove_headers:
            response["headers"].pop(header_name, None)
    return response


@pytest.fixture(scope="module", autouse=True)
def vcr_config() -> Dict[str, Any]:
    """
    vcr configuration
    """
    return {
        "filter_headers": ["authorization"],
        "before_record_response": clean_response,
    }



"""
BetterUptime error helpers.
"""
import requests

from betteruptime.typing import JSON


def parse_error_response(response: requests.Response) -> JSON:
    """
    Parse BetterUptime response to extract errors.
    """
    errors = None
    try:
        payload = response.json()
        errors = payload["errors"]
    except requests.exceptions.JSONDecodeError:
        errors = None
    except KeyError:
        errors = None

    return errors
